open Format

type statement=Var|Def|Mac

module type SIGN=
sig
  type s
  val axiom: s->(s option)
  val rule: s->s->(s option)
  val of_string: string->(s option)
  val to_string: s->string
end

module PTS=functor (MY_SIGN:SIGN)->
struct
  type t=
      Sort of MY_SIGN.s
    | Dum
    | Rel of int
    | Lam of string*t*t
    | Prod of string*t*t
    | App of t*t

  type raw=
      RDum
    | RVar of string
    | RLam of string*raw*raw
    | RProd of string*raw*raw
    | RApp of raw*raw

  type env=(string*t*(t option)) list
  
  let push (env:env) nam t=(nam,t,None)::env

  let translocate offset=
    let rec t_aux d=function
	Sort s->Sort s
      | Dum->Dum
      | Lam (nam,t1,t2)-> Lam (nam,t_aux d t1,t_aux (d+1) t2)
      | Prod (nam,t1,t2)-> Prod (nam,t_aux d t1,t_aux (d+1) t2)
      | App (t1,t2)-> App (t_aux d t1,t_aux d t2)
      | Rel x-> if x<d then Rel x else Rel (x+offset)
    in t_aux 0

  let search_var nam env=
    match MY_SIGN.of_string nam with
    Some s->Sort s
    |None->
    (let rec s_aux n=function
	([]:env)-> failwith ("Unbound named variable : "^nam)
      | (str,_,_)::q->
      	  if str=nam then Rel n else s_aux (n+1) q
    in s_aux 0 env)

  let rec search_env n=function
      ([]:env)->failwith "Looking for unbound De Bruijn." 
    | trpl::q->if n=0 then trpl else search_env (n-1) q

  let search_name n env=
    try match search_env n env with (nam,_,_)->(Some nam)
    with Failure _-> None

  let search_type n env=
    let (_,typ,_)=search_env n env in (translocate (n+1) typ)
  
  let search_body n env=
    match search_env n env with  
	(_,_,(Some bdy))->(Some (translocate (n+1) bdy))
      | _->None

  let rec unname (env:env)=function
      RDum->Dum
    | RLam (nam,t1,t2)-> 
	let tt1=unname env t1 in
	  Lam (nam,tt1,unname (push env nam tt1) t2)
    | RVar nam->search_var nam env 
    | RProd (nam,t1,t2)-> 
	let tt1=unname env t1 in
	  Prod (nam,tt1,unname (push env nam tt1) t2)
    | RApp (t1,t2)-> App (unname env t1,unname env t2)

  let subst u v= (* calcule la rduction de (app (lam u) v) *)
    let rec l_aux d=function
	Sort s->Sort s |Dum->Dum
      | Lam (nam,t1,t2)-> Lam (nam,l_aux d t1,l_aux (d+1) t2)
      | Prod (nam,t1,t2)-> Prod (nam,l_aux d t1,l_aux (d+1) t2)
      | App (t1,t2)-> App (l_aux d t1,l_aux d t2)
      | Rel x-> if x<d then Rel x
      	else if x>d then Rel (x-1) else (translocate d v)
    in l_aux 0 u

  let rec head_reduce (env:env)=function
      Rel x->(match (search_body x env) with 
      		  None->Rel x|Some bdy->head_reduce env bdy)
    | App (t1,t2)->
	(match (head_reduce env t1) with 
	     Lam (_,_,v)->head_reduce env (subst v t2)
	   |t -> App (t,t2))
    | t -> t

  let rec mac_reduce env=function
      Rel x->(match (search_body x env) with 
      	       None->Rel x|Some bdy->bdy)
    | App(t1,t2)->(match mac_reduce env t1 with
      Lam (_,Dum,u)->subst u t2
      |_-> App(t1,t2))
    | t->t

  let rec reduce (env:env)=function
      Sort s->Sort s |Dum->Dum
    | Prod (nam,t1,t2)-> 
	Prod (nam,reduce env t1,reduce (push env nam t1) t2)
    | Lam (nam,t1,t2)-> 
	Lam (nam,reduce env t1,reduce (push env nam t1) t2)
    | Rel x->(match (search_body x env) with 
      		  None->Rel x|Some bdy->reduce env bdy)
    | App (t1,t2)-> 
	(match reduce env t1 with 
      	     Lam (nam,u,v)->reduce env (subst v t2)
	   | v-> App (v,reduce env t2))
  
  let rec conv env t1 t2=
    let rec c_aux=function 
	  (Sort s1,Sort s2)->s1=s2
	| (Prod (_,t1,u1),Prod (_,t2,u2))->(c_aux (t1,t2)) && (c_aux (u1,u2))
	| (Lam (_,t1,u1),Lam (_,t2,u2))->(c_aux (t1,t2)) && (c_aux (u1,u2))
	| (Rel x,Rel y)->x=y
	| (App (t1,u1),App (t2,u2))->(c_aux (t1,t2)) && (c_aux (u1,u2)) 
	| _->false in
      c_aux (reduce env t1,reduce env t2)
		   

  let rec type_of (env:env)=function
      Dum->Dum
    | Sort s->
      	(match MY_SIGN.axiom s with 
      	     Some s0->Sort s0
	   | None->failwith "Invalid sort in a term")
    | Prod (nam,t1,t2)->
      	(match ((type_of env t1),(type_of (push env nam t1) t2)) with
	     (Sort s1,Sort s2)->
      	       (match MY_SIGN.rule s1 s2 with 
      		    Some s0->Sort s0
		  | None->failwith "Invalid product" )
      	   |_ ->failwith "Product of non-sorts")
    | Rel x->search_type x env
    | Lam (nam,t1,t2)->
      	(match t1 with Dum->Dum | _->
      	(match type_of (push env nam t1) t2 with
	  Dum->failwith "A macro was given too few arguments"
	| typ->
      	let prd=Prod (nam,t1,typ) in
      	let _=type_of env prd in prd))
    | App (t1,t2)->
      	(match type_of env t2 with 
  	 Dum->failwith "A macro was given too few arguments" 
      	|typ2->
      	(match type_of env t1 with
	Dum->type_of env (mac_reduce env (App (t1,t2)))
        | typ1->
	  ( match head_reduce env typ1 with 
      	      Prod(nam,u,v)->
      	      if conv env u typ2 then (subst v t2)
	      else failwith "Illegal application"
      	      | _->failwith "Application of a non-product")))
	

  let push_stat (env:env) (nam,rt,stat)=
    let t=unname env rt in
    match stat with
      Var->push env nam t
    | _ -> (nam,type_of env t,Some t)::env

  let nodep t=
    let rec aux n=function
	Rel x->x<>n
      | Sort _ |Dum->true
      | Lam (nam,t1,t2)|Prod (nam,t1,t2)->
	  (aux n t1) && (aux (n+1) t2)
      | App(t1,t2)->(aux n t1) && (aux n t2) in
      aux 0 t

  let rec print (env:env)=function
      Sort s->Format.print_string (MY_SIGN.to_string s)
    | Prod (nam,t1,t2)->if nam="_" || (nodep t2) then
	(printf "(";print env t1;printf "->";
	 print (push env nam t1) t2;printf ")")
      else
      	(printf "(%s:" nam;print env t1;printf ")";
	 print (push env nam t1) t2)
    | Lam (nam,t1,t2)->
      	printf "[%s:" nam;print env t1;printf "]";
	print (push env nam t1) t2
    | Dum->printf "%%"
    | Rel x-> (match search_name x env with 
      	    Some nam->printf "%s" nam
	   |None-> printf "#%d" x)
    | App (t1,t2)->
      	printf "(";print env t1;printf " ";print env t2;printf ")"

 let rec print_mac (env:env)=function
  Lam (nam,Dum,t2)->printf "%s " nam;print_mac (push env nam (Rel(-1))) t2
  |t->printf "= ";print env t

  let print_stat (env:env) (nam,typ,bdy)=
    match typ,bdy with 
    Dum,Some t1->printf "macro %s " nam;print_mac env t1;printf ".\n"
  | t1,None->printf "axiom %s : " nam;print env t1;printf ".\n"
  | t1,Some t2-> 
      printf "define %s : " nam;print env t1;printf " = ";
      print env t2;printf ".\n"

  let rec dump=function
    ([]:env)->()
  |stat::env -> dump env;print_stat env stat

  let read_prog=List.fold_left push_stat []

end

module MY_PTS=PTS(Mksys)

